from abml_dataclass import Abml_Registry

# from abml_helpers import cprint
from abaqus import mdb
from abaqusConstants import ANALYSIS, DEFAULT, OFF
import os
from shutil import copy, move


@Abml_Registry.register("jobs")
class Abml_Job:
    def __init__(self, name, model, **kwargs):  # noqa
        self.name = str(name)
        self.model = model
        self.kwargs = kwargs
        self.create()

    def create(self):
        subroutine_flag = self.kwargs.get("subroutine", False)
        if subroutine_flag:
            subroutine_path = os.path.abspath("subs.for")
        else:
            subroutine_path = None

        kwargs = {
            "name": self.name,
            "model": self.model.name,
            "description": self.kwargs.get("description", ""),
            "type": ANALYSIS,
            "userSubroutine": subroutine_path,
            "numCpus": self.kwargs.get("cpus", 1),
            "numGPUs": self.kwargs.get("gpus", 0),
            "multiprocessingMode": DEFAULT,
        }

        mdb.Job(**kwargs)

    def write_input(self):
        mdb.jobs[self.name].writeInput(consistencyChecking=OFF)

    def write_and_copy_input_to_path(self, path):
        self.write_input()
        filename = "{}.inp".format(self.name)
        copy(filename, os.path.join(path, filename))

    def write_and_move_input_to_path(self, path):
        self.write_input()
        filename = "{}.inp".format(self.name)
        move(filename, os.path.join(path, filename))



# from abaqus import Mdb
from json import JSONEncoder, dumps
from numpy import ndarray, empty, array, vstack
import __main__

import argparse
import os
from io import open

from abaqus import Mdb
from abaqus import *  # noqa # type: ignore[import-star]
from abml import abml_helpers


parser = argparse.ArgumentParser()
parser.add_argument("--filename")
parser.add_argument("--calls", "-c", nargs="+", help="sequence calls")

args, _ = parser.parse_known_args()

os.environ["ABAQUS_NO_MPI"] = "true"


class NumpyArrayEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, ndarray):
            return o.round(10).tolist()
        if isinstance(o, dict):
            new_dict = {}
            for k, v in o.items():
                if isinstance(v, ndarray):
                    new_dict[k] = v.round(10).tolist()
                elif isinstance(v, dict):
                    new_dict[k] = self.default(v)
                else:
                    new_dict[k] = v
            return new_dict
        return JSONEncoder.default(self, o)


def extract_hierarchy(input_str):
    hierarchy = []
    current = ""
    is_quoted = False

    for char in input_str:
        if char == "'":
            is_quoted = not is_quoted
        elif is_quoted:
            current += char
        elif char == "[":
            if current:
                hierarchy.append(current)
            current = ""
        elif char == "]":
            if current:
                hierarchy.append(current)
            current = ""
        elif char == ".":
            if current:
                hierarchy.append(current)
            current = ""
        else:
            current += char

    if current:
        hierarchy.append(current)

    return hierarchy


def get_vertices(repositories, vertices):
    repositories = array([repositories]).flatten()

    vertices_extract = empty((len(repositories), 3))
    for repository in repositories:
        for vertex_index in repository.getVertices():
            vertices_extract = vstack((vertices_extract, vertices[vertex_index].pointOn))

    return vertices_extract


def get_nested_value(d, *keys):
    """
    Get the value associated with the sequence of keys in the nested dictionary or object d.

    Args:
        d (dict or object): The nested dictionary or object.
        keys (tuple): The sequence of keys.

    Returns:
        The value associated with the sequence of keys in the nested dictionary or object.
    """
    for key in keys:
        if not isinstance(d, (dict, object)):
            return None
        if isinstance(d, dict):
            d = d.get(key)
            if isinstance(d, dict):
                for val in d.values():
                    if not isinstance(val, object):
                        return None
        elif isiterable(d):
            try:
                d = getattr(d, key)
            except AttributeError:
                d = d[int(key)]
        else:
            try:
                d = getattr(d, key)
            except AttributeError:
                d = d[key]

    return d


def isiterable(obj):
    try:
        iter(obj)
        return True
    except TypeError:
        return False


# def load_data_from_jnl(jnl, model_name="Model-1"):
#     mdb = Mdb()
#     execfile(jnl, __main__.__dict__)  # noqa

#     m = mdb.models[model_name]
#     p = m.parts

#     data = {}
#     data["parts"] = {}
#     for part_key, part in p.items():
#         data["parts"][part_key] = {}
#         data_part = data["parts"][part_key]
#         data_part["vertices"] = {"cells": get_vertices(part.vertices)}
#         data_part["sets"] = get_part_sets(part.sets)

#     return data


def drop_strings(input_list, strings_to_drop):
    return [elem for elem in input_list if elem not in strings_to_drop]


def del_files():
    for file in os.listdir("."):
        if ".rpy" in file or ".rec" in file:
            try:
                os.remove(file)
            except WindowsError:
                pass


if __name__ == "__main__":
    mdb = Mdb()
    execfile(args.filename, __main__.__dict__)  # noqa

    m = mdb.models
    abml_helpers.print_to_console(args.calls)
    for call in args.calls:
        call = extract_hierarchy(call)

        repository_call = drop_strings(call, ["mdb", "models"])
        repository = get_nested_value(m, *repository_call)

        vertices_call = repository_call[:-1]
        vertices_call.append("vertices")
        vertices = get_nested_value(m, *vertices_call)
        vertices_export = get_vertices(repository, vertices)

        abml_helpers.print_to_console(repository, "#+#")
        abml_helpers.print_to_console(vertices_export, "#+#")

        with open(args.filename.replace(".jnl", ".json"), mode="w", encoding="utf8") as f:
            json_str = dumps(vertices_export, ensure_ascii=False, cls=NumpyArrayEncoder, indent=2)

            if isinstance(json_str, str):
                json_str = json_str.decode("utf-8")

            f.write(json_str)
            # dump(vertices, f, cls=NumpyArrayEncoder, indent=2, ensure_ascii=False)

    del_files()
    # todo: jnl-gen -k Model-1 -k parts -k Part-1 -k faces -k [0]

    # mdb.saveAs(pathName="my_model.cae")

    # with open(args.filename.replace(".jnl", ".json"), mode="w", encoding="utf-8") as f:
    #     dump(data, f, cls=NumpyArrayEncoder)



from abml_dataclass import Abml_Registry, to_object_key_based_type
from abaqusConstants import FRICTIONLESS, PENALTY, ISOTROPIC, FRACTION
from abaqusConstants import HARD, DEFAULT
from abaqusConstants import ALL_NODES
from abaqusConstants import ON, OFF


@Abml_Registry.register("interaction_properties")
class Abml_Interaction_Prop:
    def __init__(self, name, type, behaviors, model, **kwargs):  # noqa
        self.name = name
        self.type = type
        self.model = model

        self.create()
        self.behaviors = to_object_key_based_type(behaviors, interaction=self)

    def create(self):
        self.model.m.ContactProperty(self.name)


@Abml_Registry.register("tangential")
class Abml_Tangential_Behavior:
    def __init__(self, formulation, interaction, **kwargs):
        self.formulation = formulation
        self.interaction = interaction
        self.interaction_prop = interaction.model.m.interactionProperties[interaction.name]
        self.formulation_map = {"frictionless": FRICTIONLESS, "penalty": PENALTY}
        self.create_map = {
            "frictionless": self.create_frictionless,
            "penalty": self.create_penalty,
        }
        self.kwargs = kwargs
        self.create()

    def create(self):
        self.create_map[self.formulation]()

    def create_frictionless(self):
        kwargs = {"formulation": self.formulation_map[self.formulation]}
        self.interaction_prop.TangentialBehavior(**kwargs)

    def create_penalty(self):
        kwargs = {
            "formulation": self.formulation_map[self.formulation],
            "directionality": ISOTROPIC,
            "slipRateDependency": OFF,
            "pressureDependency": OFF,
            "temperatureDependency": OFF,
            "dependencies": 0,
            "table": ((self.kwargs["friction_coeff"],),),
            "shearStressLimit": None,
            "maximumElasticSlip": FRACTION,
            "fraction": 0.005,
            "elasticSlipStiffness": None,
        }

        self.interaction_prop.TangentialBehavior(**kwargs)


@Abml_Registry.register("normal")
class Abml_Normal_Behavior:
    def __init__(self, overclosure, constraint, seperation, interaction):
        self.overclosure = overclosure
        self.constraint = constraint
        self.seperation = seperation
        self.interaction = interaction
        self.interaction_prop = interaction.model.m.interactionProperties[interaction.name]
        self.overclosure_map = {"hard_contact": HARD}
        self.constraint_map = {"default": DEFAULT}

        self.create()

    def create(self):
        kwargs = {
            "pressureOverclosure": self.overclosure_map[self.overclosure],
            "allowSeparation": self.seperation,
            "constraintEnforcementMethod": self.constraint_map[self.constraint],
        }

        self.interaction_prop.NormalBehavior(**kwargs)


@Abml_Registry.register("cohesive")
class Abml_Cohesive_Behavior:
    def __init__(self, interaction, **kwargs):
        self.interaction = interaction
        self.interaction_prop = interaction.model.m.interactionProperties[interaction.name]
        self.kwargs = kwargs
        self.create()

    def create(self):
        kwargs = {
            "eligibility": ALL_NODES,
            "defaultPenalties": OFF,
            "table": ((self.kwargs["knn"], self.kwargs["kss"], self.kwargs["ktt"]),),
        }

        self.interaction_prop.CohesiveBehavior(**kwargs)


@Abml_Registry.register("damage")
class Abml_Damage_Behavior:
    def __init__(self, interaction, **kwargs):
        self.interaction = interaction
        self.interaction_prop = interaction.model.m.interactionProperties[interaction.name]
        self.kwargs = kwargs
        self.create()

    def create(self):
        kwargs = {
            "initTable": ((self.kwargs["normal_only"], self.kwargs["shear_1_only"], self.kwargs["shear_2_only"]),),
            "useEvolution": ON,
            "evolTable": ((self.kwargs["total_plastic_displacement"],),),
        }

        self.interaction_prop.Damage(**kwargs)



# import logging

# debug_level = logging.CRITICAL

# parts_logger = logging.getLogger("abml.parser")
# parts_logger.addHandler(logging.FileHandler("parser.log", encoding="utf-8", mode="w"))
# parts_logger.setLevel(debug_level)
# parts_logger.propagate = False

# parts_logger = logging.getLogger("abml.abml_parts")
# parts_logger.addHandler(logging.FileHandler("parts.log", encoding="utf-8", mode="w"))
# parts_logger.setLevel(debug_level)
# parts_logger.propagate = False

# assembly_logger = logging.getLogger("abml.abml_assembly")
# assembly_logger.addHandler(logging.FileHandler("assembly.log", encoding="utf-8", mode="w"))
# assembly_logger.setLevel(debug_level)
# assembly_logger.propagate = False

# sections_logger = logging.getLogger("abml.abml_sections")
# sections_logger.addHandler(logging.FileHandler("sections.log", encoding="utf-8", mode="w"))
# sections_logger.setLevel(debug_level)
# sections_logger.propagate = False

# materials_logger = logging.getLogger("abml.abml_materials")
# materials_logger.addHandler(logging.FileHandler("materials.log", encoding="utf-8", mode="w"))
# materials_logger.setLevel(debug_level)
# materials_logger.propagate = False

# helpers_logger = logging.getLogger("abml.abml_helpers")
# helpers_logger.addHandler(logging.FileHandler("helpers.log", encoding="utf-8", mode="w"))
# helpers_logger.setLevel(debug_level)
# helpers_logger.propagate = False

# helpers_logger = logging.getLogger("abml.abml_steps")
# helpers_logger.addHandler(logging.FileHandler("steps.log", encoding="utf-8", mode="w"))
# helpers_logger.setLevel(debug_level)
# helpers_logger.propagate = False



import types
import json
import inspect
from collections import OrderedDict
from collections import defaultdict
import sys
import yaml
from io import open
from abaqus import mdb
from abaqusConstants import STANDARD_EXPLICIT
from abml import abml_helpers
import interaction


class Abml_Registry:
    registry = defaultdict(lambda: OrderedDict)

    @classmethod
    def register(cls, key):
        def decorator(dataclass):
            cls.registry[key] = dataclass
            return dataclass

        return decorator


@Abml_Registry.register("cae")
class Abml_Cae:
    def __init__(self, models, **kwargs):
        self.models = to_object_key_mode(models["models"], type_="models")

    def save_cae(self, filename):
        mdb.saveAs(pathName=filename)


@Abml_Registry.register("models")
class Abml_Model:
    def __init__(self, name, **kwargs):
        self.name = name
        self.create()
        self.m, self.a, self.p = abml_helpers.get_objects(self.name)

        materials_list = [value for key, value in kwargs.items() if "materials" in key]
        for materials in materials_list:
            if materials is not None:
                materials = OrderedDict(sorted(materials.items()))
                self.materials = to_object_key_mode(materials, "materials", model=self)

        sections_list = [value for key, value in kwargs.items() if "sections" in key]
        for sections in sections_list:
            if sections is not None:
                sections = OrderedDict(sorted(sections.items()))
                self.materials = to_object_key_mode(sections, "sections", model=self)

        parts_list = [value for key, value in kwargs.items() if "parts" in key]
        for parts in parts_list:
            if parts is not None:
                parts = OrderedDict(sorted(parts.items()))
                self.parts = to_object_key_mode(parts, "parts", model=self)

        assembly_list = [value for key, value in kwargs.items() if "assembly" in key]
        for assembly in assembly_list:
            if assembly is not None:
                self.assembly = to_object_hierachy_mode(assembly, "assembly", model=self)

        steps_list = [value for key, value in kwargs.items() if "steps" in key]
        for steps in steps_list:
            if steps is not None:
                steps = OrderedDict(sorted(steps.items()))
                self.interaction_props = to_object_key_mode(steps, "steps", model=self)

        interaction_props_list = [value for key, value in kwargs.items() if "interaction_props" in key]
        for interaction_props in interaction_props_list:
            if interaction_props is not None:
                interaction_props = OrderedDict(sorted(interaction_props.items()))
                self.interaction_props = to_object_key_mode(interaction_props, "interaction_properties", model=self)

        interactions_list = [value for key, value in kwargs.items() if "interactions" in key]
        for interactions in interactions_list:
            if interactions is not None:
                interactions = OrderedDict(sorted(interactions.items()))
                self.interactions = to_object_key_mode(interactions, "interactions", model=self)

        constraints_list = [value for key, value in kwargs.items() if "constraints" in key]
        for constraints in constraints_list:
            if constraints is not None:
                constraints = OrderedDict(sorted(constraints.items()))
                self.constraints = to_object_key_mode(constraints, "constraints", model=self)

        bcs_list = [value for key, value in kwargs.items() if "bcs" in key]
        for bcs in bcs_list:
            if bcs is not None:
                bcs = OrderedDict(sorted(bcs.items()))
                self.bcs = to_object_key_mode(bcs, "bcs", model=self)

        loads_list = [value for key, value in kwargs.items() if "loads" in key]
        for loads in loads_list:
            if loads is not None:
                loads = OrderedDict(sorted(loads.items()))
                self.loads = to_object_key_mode(loads, "loads", model=self)

        jobs_list = [value for key, value in kwargs.items() if "jobs" in key]
        for jobs in jobs_list:
            if jobs is not None:
                jobs = OrderedDict(sorted(jobs.items()))
                self.jobs = to_object_key_mode(jobs, "jobs", model=self)

    def create(self):
        mdb.Model(modelType=STANDARD_EXPLICIT, name=self.name)

        if "Model-1" in mdb.models.keys():
            del mdb.models["Model-1"]


def data_to_object(data, type_="models", key_based=True, **kwargs):
    if data is not None:
        object_ = {}
        if isinstance(data, dict):
            object_ = {}
            if key_based:
                for key, value in data.items():
                    if isinstance(value, dict):
                        value.update(kwargs)
                        object_[key] = Abml_Registry.registry[type_](name=key, **value)
                    elif isinstance(value, (list)):
                        object_[key] = Abml_Registry.registry[type_](name=key, *value)
            else:
                data.update(kwargs)
                object_ = Abml_Registry.registry[type_](**data)
        elif isinstance(data, list):
            object_ = []
            for value in data:
                value.update(kwargs)
                object_.append(Abml_Registry.registry[type_](**value))

        return object_


def to_object_key_mode(data, type_, **kwargs):
    object_ = {}
    for key, value in data.items():
        value.update(kwargs)
        object_[key] = Abml_Registry.registry[type_](name=key, **value)

    return object_


def to_object_option_mode(data, type_, **kwargs):
    data.update(kwargs)
    object_ = Abml_Registry.registry[type_](**data)

    return object_


def to_object_key_based_type(data, **kwargs):
    object_ = {}
    for key, value in data.items():
        value.update(kwargs)
        object_[key] = Abml_Registry.registry[key](**value)

    return object_


def to_object_cmd_mode(data, type_, **kwargs):
    object_ = []
    for cmd in data:
        value = cmd.values()[0]
        key = cmd.keys()[0]
        value.update(kwargs)
        object_.append(Abml_Registry.registry[type_](cmd=key, **value))

    return object_


def to_object_hierachy_mode(data, type_, **kwargs):
    # object_ = []
    # data.update(kwargs)
    # object_.append(Abml_Registry.registry[type_](**data))

    data.update(kwargs)
    object_ = Abml_Registry.registry[type_](**data)

    return object_



from abml_dataclass import Abml_Registry
from regionToolset import Region
from abaqusConstants import FINITE, SMALL


@Abml_Registry.register("interactions")
class Abml_Interaction:
    def __init__(self, name, type, model, **kwargs):  # noqa
        self.name = name
        self.type = type
        self.model = model
        self.kwargs = kwargs

        self.map = {
            "surface_to_surface": self.surface_to_surface,
            "elastic_foundation": self.elastic_foundation,
        }
        self.sliding_map = {"finite": FINITE, "small": SMALL}
        self.create()

    def create(self):
        self.map[self.type]()

    def surface_to_surface(self):
        kwargs = {
            "name": self.name,
            "master": Region(side1Faces=self.model.assembly.get_seq_from_surface_list(self.kwargs["master"])),
            "slave": Region(side1Faces=self.model.assembly.get_seq_from_surface_list(self.kwargs["slave"])),
            "createStepName": self.kwargs.get("step", "Initial"),
            "sliding": self.sliding_map[self.kwargs.get("sliding", "finite").lower()],
            "interactionProperty": self.kwargs["interaction_prop"],
        }

        self.model.m.SurfaceToSurfaceContactStd(**kwargs)

    def elastic_foundation(self):
        try:
            surface_region = Region(side1Faces=self.model.assembly.get_seq_from_surface_list(self.kwargs["faces"]))
        except TypeError as e:
            error_msg = """
            model_name: {model_name}
            interaction_name: {name}
            surfaces: {surfaces}
            error: {e}
            """.format(
                model_name=self.model.name,
                name=self.name,
                e=e,
                surfaces=self.kwargs["faces"],
            )

            raise TypeError(error_msg)
        kwargs = {
            "name": self.name,
            "surface": surface_region,
            "createStepName": "Initial",
            "stiffness": self.kwargs["stiffness"],
        }

        self.model.m.ElasticFoundation(**kwargs)



from abml_dataclass import Abml_Registry
from abml_helpers import Abml_Helpers
from abml_helpers import cprint


@Abml_Registry.register("sections")
class Abml_Section(Abml_Helpers):
    def __init__(self, model, name, category, type, material, **kwargs):  # noqa
        self.model = model
        self.name = name
        self.category = category
        self.type = type
        self.material = material
        self.map = {}

        self.section_map = {("homogeneous", "solid"): self.homogeneous_solid}
        self.create()

    def homogeneous_solid(self):
        kwargs = {
            "name": self.name,
            "material": self.material,
        }
        self.model.m.HomogeneousSolidSection(**kwargs)

    def create(self):
        self.section_map[(self.type, self.category)]()



from abml_dataclass import Abml_Registry
from abml_helpers import Abml_Helpers, cprint
from abaqusConstants import USER_DEFINED, UNSET, UNIFORM, SET, UNSET
from regionToolset import Region


@Abml_Registry.register("bcs")
class Abml_Bc(Abml_Helpers):
    def __init__(self, name, model, type, **kwargs):
        self.name = name
        self.step = kwargs.get("step", "Initial")
        self.model = model
        self.type = type
        self.faces = self.convert_position_list(kwargs.get("faces", []))
        self.edges = self.convert_position_list(kwargs.get("edges", []))
        self.cells = self.convert_position_list(kwargs.get("cells", []))
        self.kwargs = kwargs

        self.map = {
            "x_symm": self.x_symm,
            "y_symm": self.y_symm,
            "z_symm": self.z_symm,
            "disp_rot": self.disp_rot,
        }

        self.set_map = {
            "set": SET,
            "unset": UNSET,
            1: SET,
            0: UNSET,
        }
        self.create()

    def create(self):
        self.map[self.type]()

    def x_symm(self):
        region = Region(faces=self.model.assembly.get_seq_from_surface_list(self.kwargs["faces"]))
        self.model.m.XsymmBC(name=self.name, createStepName=self.step, region=region, localCsys=None)

    def y_symm(self):
        region = Region(faces=self.model.assembly.get_seq_from_surface_list(self.kwargs["faces"]))
        self.model.m.YsymmBC(name=self.name, createStepName=self.step, region=region, localCsys=None)

    def z_symm(self):
        region = Region(faces=self.model.assembly.get_seq_from_surface_list(self.kwargs["faces"]))
        self.model.m.ZsymmBC(name=self.name, createStepName=self.step, region=region, localCsys=None)

    def disp_rot(self):
        region = Region(faces=self.model.assembly.get_seq_from_surface_list(self.kwargs["faces"]))
        self.model.m.DisplacementBC(
            name=self.name,
            createStepName=self.step,
            region=region,
            u1=self.set_map[self.kwargs.get("u1", 0)],
            u2=self.set_map[self.kwargs.get("u2", 0)],
            u3=self.set_map[self.kwargs.get("u3", 0)],
            ur1=self.set_map[self.kwargs.get("ur1", 0)],
            ur2=self.set_map[self.kwargs.get("ur2", 0)],
            ur3=self.set_map[self.kwargs.get("ur3", 0)],
            amplitude=UNSET,
            distributionType=UNIFORM,
            fieldName="",
            localCsys=None,
        )


# symm:
#     xsymm:
#         faces: [platte1.E]
# test2:
#     disp_rot:
#         faces: [platte2.E]
#         u1: 1
#         u2: 1
#         u3: 1

# self.model = part.model
# self.faces = self.convert_position_list(faces)
# self.edges = self.convert_position_list(edges)
# self.cells = self.convert_position_list(cells)
# self.kwargs = kwargs
# self.partition_map = {"plane": self.create_face_plane}

# self.create()

# def create(self):
# self.partition_map[self.method]()

# def create_face_plane(self):
# part_ = self.model.p[self.part.name]
# planes = array([self.kwargs.get("planes", [])]).flatten()

# for plane in planes:
#     id_ = part_.features[plane].id

#     if self.faces is not None:
#         face_seq = self.part.geometry.get_sequence_from_list(self.faces, "face")
#         if face_seq is not None:
#             part_.PartitionFaceByDatumPlane(datumPlane=part_.datums[id_], faces=face_seq)

#     if self.cells is not None:
#         cell_seq = self.part.geometry.get_sequence_from_list(self.cells, "cell")
#         if cell_seq is not None:
#             part_.PartitionCellByDatumPlane(datumPlane=part_.datums[id_], cells=cell_seq)

#     if self.edges is not None:
#         edge_seq = self.part.geometry.get_sequence_from_list(self.edges, "edge")
#         if edge_seq is not None:
#             part_.PartitionEdgeByDatumPlane(datumPlane=part_.datums[id_], edges=edge_seq)



from abml_helpers import Abml_Helpers
import abml_helpers
from abml_parts import Abml_Part
from abml_assembly import Abml_Assembly
from abml_dataclass import Abml_Registry
from abml_dataclass import Abml_Model
from abaqusConstants import FREE, FINER, FIXED, STRUCTURED
from abaqusConstants import TET, HEX
from abaqusConstants import STANDARD, DEFAULT, OFF
from abaqusConstants import C3D8, C3D8R, C3D6, C3D4, C3D20, C3D20R, C3D15, C3D10
from regionToolset import Region
import mesh


@Abml_Registry.register("mesh")
class Abml_Mesh_Command(Abml_Helpers):
    def __init__(self, cmd, part=None, assembly=None, **kwargs):
        self.part = part
        self.assembly = assembly
        self.cmd = cmd
        self.kwargs = kwargs

        if part is not None:
            self.abaqus_object = self.part.geometry  # noqa
            self.model = part.model
        elif assembly is not None:
            self.abaqus_object = self.assembly
            self.model = assembly.model

        self.map = {
            "seed_edge_by_size": self.seed_edge_size,
            "seed_edge_by_number": self.seed_edge_by_number,
            "mesh_controls": self.mesh_controls,
            "element_type": self.element_type,
        }

        self.elem_shape_map = {"tet": TET, "hex": HEX}
        self.technique_map = {"free": FREE, "structured": STRUCTURED}
        self.constraint_map = {"free": FREE, "finer": FINER, "fixed": FIXED}
        self.element_types_map = {
            ("linear", True): self._element_type_linear_reduced,
            ("quadratic", True): self._element_type_quadratic_reduced,
            ("linear", False): self._element_type_linear,
            ("quadratic", False): self._element_type_quadratic,
        }
        self.create()

    def create(self):
        self.map.get(self.cmd, lambda: None)()

    def seed_edge_size(self):
        # edges = abaqus_object.get_sequence_from_list(self.convert_position_list(self.kwargs["edges"]), "edge")
        kwargs = {
            "edges": 0,
            "size": self.kwargs["size"],
            "constraint": self.constraint_map[self.kwargs.get("constraint", "finer").lower()],
            "deviationFactor": self.kwargs.get("deviationFactor", None),
            "minSizeFactor": self.kwargs.get("minSizeFactor", None),
        }
        self.abaqus_object.seedEdgeBySize(**kwargs)

    def seed_edge_by_number(self):
        edges_list = self.convert_position_list(self.kwargs.get("edges", None))
        set_list = self.convert_position_list(self.kwargs.get("sets", None))
        edge_seq = None
        if edges_list is not None:
            seq = self.abaqus_object.get_sequence_from_list(edges_list, "edge")
            edge_seq = seq
        if set_list is not None:
            seq = self.abaqus_object.get_sequence_from_list(set_list, "set_edges")
            if edge_seq is None:
                edge_seq = seq
            else:
                edge_seq += seq
        kwargs = {
            "edges": edge_seq,
            "number": self.kwargs["number"],
            "constraint": self.constraint_map[self.kwargs.get("constraint", "finer").lower()],
        }

        self.model.p[self.part.name].seedEdgeByNumber(**kwargs)

    def mesh_controls(self):
        cells = self.abaqus_object.get_sequence_from_list(self.convert_position_list(self.kwargs["cells"]), "cell")

        kwargs = {
            "regions": cells,
            "elemShape": self.elem_shape_map[self.kwargs["shape"]],
            "technique": self.technique_map[self.kwargs.get("technique", "free")],
        }
        self.model.p[self.part.name].setMeshControls(**kwargs)

    def element_type(self):
        cell_list = self.convert_position_list(self.kwargs["cells"])

        if cell_list is not None:
            cells = self.abaqus_object.get_sequence_from_list(cell_list, "cell")

            elemType1, elemType2, elemType3 = self.element_types_map[
                (self.kwargs.get("order"), self.kwargs.get("reduced"))
            ]()

            kwargs = {
                "regions": Region(cells=cells),
                "elemTypes": (elemType1, elemType2, elemType3),
            }
            # self.model.p[self.part.name].setMeshControls(**kwargs)

            self.model.p[self.part.name].setElementType(**kwargs)

    def _element_type_linear_reduced(self):
        elemType1 = mesh.ElemType(
            elemCode=C3D8R, elemLibrary=STANDARD, secondOrderAccuracy=OFF, distortionControl=DEFAULT
        )
        elemType2 = mesh.ElemType(
            elemCode=C3D6, elemLibrary=STANDARD, secondOrderAccuracy=OFF, distortionControl=DEFAULT
        )
        elemType3 = mesh.ElemType(
            elemCode=C3D4, elemLibrary=STANDARD, secondOrderAccuracy=OFF, distortionControl=DEFAULT
        )

        return elemType1, elemType2, elemType3

    def _element_type_linear(self):
        elemType1 = mesh.ElemType(
            elemCode=C3D8, elemLibrary=STANDARD, secondOrderAccuracy=OFF, distortionControl=DEFAULT
        )
        elemType2 = mesh.ElemType(
            elemCode=C3D6, elemLibrary=STANDARD, secondOrderAccuracy=OFF, distortionControl=DEFAULT
        )
        elemType3 = mesh.ElemType(
            elemCode=C3D4, elemLibrary=STANDARD, secondOrderAccuracy=OFF, distortionControl=DEFAULT
        )

        return elemType1, elemType2, elemType3

    def _element_type_quadratic_reduced(self):
        elemType1 = mesh.ElemType(elemCode=C3D20R, elemLibrary=STANDARD)
        elemType2 = mesh.ElemType(elemCode=C3D15, elemLibrary=STANDARD)
        elemType3 = mesh.ElemType(
            elemCode=C3D10, elemLibrary=STANDARD, secondOrderAccuracy=OFF, distortionControl=DEFAULT
        )

        return elemType1, elemType2, elemType3

    def _element_type_quadratic(self):
        elemType1 = mesh.ElemType(elemCode=C3D20, elemLibrary=STANDARD)
        elemType2 = mesh.ElemType(elemCode=C3D15, elemLibrary=STANDARD)
        elemType3 = mesh.ElemType(
            elemCode=C3D10, elemLibrary=STANDARD, secondOrderAccuracy=OFF, distortionControl=DEFAULT
        )

        return elemType1, elemType2, elemType3



from regionToolset import Region
from abml_dataclass import Abml_Registry, to_object_key_mode
from abml_helpers import Abml_Helpers, cprint
from abaqusConstants import ON, OFF, CARTESIAN
from numpy import array


@Abml_Registry.register("assembly")
class Abml_Assembly(Abml_Helpers):
    def __init__(self, model, instances=None, surfaces=None, sets=None, **kwargs):  # noqa
        self.model = model
        if instances is not None:
            self.instances = to_object_key_mode(instances, "instances", model=self.model)
        if surfaces is not None:
            self.surfaces = to_object_key_mode(surfaces, "assembly_surfaces", model=model, assembly=self)
        if sets is not None:
            self.sets = to_object_key_mode(sets, "assembly_sets", model=model, assembly=self)

    def get_seq_from_surface_list(self, surfaces):
        surfaces = self.convert_position_list(surfaces)
        instances = self.instances
        face_seq = None

        surface_keys = set(surfaces) & set(self.model.a.surfaces.keys())
        for surface_key in surface_keys:
            seq = self.model.a.surfaces[surface_key]
            if face_seq is None:
                face_seq = seq
            else:
                face_seq += seq

        faces = (set(surfaces) ^ set(self.model.a.surfaces.keys())) & set(surfaces)
        for instance in instances.values():
            seq = instance.get_sequence_from_list(faces, "face")
            if face_seq is None:
                face_seq = seq
            elif seq is not None:
                face_seq += seq
        return face_seq

    def get_seq_from_edge_list(self, edges):
        edges = self.convert_position_list(edges)
        instances = self.instances
        edge_seq = None

        edge_keys = set(edges) & set(self.model.a.sets.keys())
        for edge_key in edge_keys:
            seq = self.model.a.sets[edge_key].edges
            if edge_seq is None:
                edge_seq = seq
            else:
                edge_seq += seq

        edges_ = (set(edges) ^ set(self.model.a.sets.keys())) & set(edges)
        for instance in instances.values():
            seq = instance.get_sequence_from_list(edges_, "edge")
            if edge_seq is None:
                edge_seq = seq
            elif seq is not None:
                edge_seq += seq
        return edge_seq

    def get_seq_from_cell_list(self, cells):
        cells = self.convert_position_list(cells)
        instances = self.instances
        cell_seq = None

        cell_keys = set(cells) & set(self.model.a.sets.keys())
        for cell_key in cell_keys:
            seq = self.model.a.sets[cell_key].cells
            if cell_seq is None:
                cell_seq = seq
            else:
                cell_seq += seq

        cell_ = (set(cells) ^ set(self.model.a.sets.keys())) & set(cells)
        for instance in instances.values():
            seq = instance.get_sequence_from_list(cell_, "cell")
            if cell_seq is None:
                cell_seq = seq
            elif seq is not None:
                cell_seq += seq
        return cell_seq


@Abml_Registry.register("instances")
class Abml_Instance:
    def __init__(self, name, model, part, **kwargs):
        self.model = model
        self.name = name
        self.part = part
        self.dependent = kwargs.get("dependent", True)
        self.translate = kwargs.get("translate", [0, 0, 0])
        self.rotate = kwargs.get("rotate", [])
        self.create()

    def create(self):
        dep_dict = {False: OFF, True: ON}

        self.model.a.DatumCsysByDefault(CARTESIAN)
        self.model.a.Instance(dependent=dep_dict[self.dependent], name=self.name, part=self.model.p[self.part])

        if len(self.rotate) != 0:
            for rot_param in self.rotate:
                angle = rot_param.get("angle")
                axisDirection = rot_param.get("axisDirection")
                axisPoint = rot_param.get("axisPoint")
                self.model.a.rotate(
                    angle=angle,
                    axisDirection=axisDirection,
                    axisPoint=axisPoint,
                    instanceList=(self.name,),
                )

        self.model.a.instances[self.name].translate(vector=self.translate)

    def get_sequence_from_list(self, elements, type_):
        sequence_map = {
            "cell": self.get_cell_sequence,
            "face": self.get_face_sequence,
            "edge": self.get_edge_sequence,
        }

        type_seq = None
        for element in elements:
            if len(element) != 0:
                seq = sequence_map[type_](element)
                if type_seq is None:
                    type_seq = seq
                elif seq is not None:
                    type_seq += seq
        return type_seq

    def get_face_sequence(self, element):
        instance = self.model.a.instances[self.name]
        element = array([element]).flatten()
        seq = None
        if element.shape == (3,) and isinstance(element[0], (float, int)):  # findAt
            face = instance.faces.findAt(element.astype(float))  # noqa
            if face is not None:
                seq = Region(side1Faces=instance.faces[face.index : face.index + 1]).side1Faces
        elif element.shape == (6,) and isinstance(element[0], (float, int)):  # getByBoundingBox
            faces = instance.faces.getByBoundingBox(*element.astype(float))
            if len(faces) != 0:
                seq = faces
        elif isinstance(element.item(0), str):  # getByBoundingBox str
            instance_name, surface_name = element.item(0).split(".")
            if instance_name == self.name:
                faces = self.model.a.instances[self.name].surfaces[surface_name].faces
                seq = faces
        return seq

    def get_cell_sequence(self, element):
        instance = self.model.a.instances[self.name]
        element = array([element]).flatten()
        seq = None
        if element.shape == (3,) and isinstance(element[0], (float, int)):  # findAt
            cell = instance.cells.findAt(element.astype(float))
            if cell is not None:
                seq = Region(cells=instance.cells[cell.index : cell.index + 1]).cells
        elif element.shape == (6,) and isinstance(element[0], (float, int)):  # getByBoundingBox
            cells = instance.cells.getByBoundingBox(*element.astype(float))
            if len(cells) != 0:
                seq = cells
        return seq

    def get_edge_sequence(self, element):
        instance = self.model.a.instances[self.name]
        element = array([element]).flatten()
        seq = None
        if element.shape == (3,) and isinstance(element[0], (float, int)):  # findAt
            edge = instance.edges.findAt(element.astype(float))
            if isinstance(edge, type(None)):
                seq = None
            else:
                seq = Region(edges=instance.edges[edge.index : edge.index + 1]).edges
        elif element.shape == (6,) and isinstance(element[0], (float, int)):  # getByBoundingBox
            edges = seq = instance.edges.getByBoundingBox(*element.astype(float))
            if len(edges) != 0:
                seq = edges
        return seq


@Abml_Registry.register("assembly_surfaces")
class Abml_Assembly_Surface(Abml_Helpers):
    def __init__(self, name, model, assembly, faces=None):
        self.name = name
        self.model = model
        self.faces = self.convert_position_list(faces)
        self.assembly = assembly

        self.create_surfaces()

    def create_surfaces(self):
        instances = self.assembly.instances

        seq = None
        for face in self.faces:
            for instance in instances.values():
                face_seq = instance.get_face_sequence(face)
                if face_seq is not None:
                    if seq is None:
                        seq = face_seq
                    else:
                        seq += face_seq

        self.model.a.Surface(side1Faces=seq, name=self.name)
        del seq


@Abml_Registry.register("assembly_sets")
class Abml_Assembly_Set(Abml_Helpers):
    def __init__(self, name, model, assembly, **kwargs):
        self.name = name
        self.model = model
        self.faces = self.convert_position_list(kwargs.get("faces", None))
        self.edges = self.convert_position_list(kwargs.get("edges", None))
        self.cells = self.convert_position_list(kwargs.get("cells", None))
        self.assembly = assembly

        self.create()

    def create(self):
        assembly = self.model.a

        kwargs = {}
        face_seq = None
        instances = self.assembly.instances

        face_seq = None
        for instance in instances.values():
            seq = instance.get_sequence_from_list(self.faces, "face")
            if face_seq is None:
                face_seq = seq
            elif seq is not None:
                face_seq += seq

        if face_seq is not None:
            kwargs["faces"] = face_seq

        cell_seq = None
        for instance in instances.values():
            seq = instance.get_sequence_from_list(self.cells, "cell")
            if cell_seq is None:
                cell_seq = seq
            elif seq is not None:
                cell_seq += seq
        if cell_seq is not None:
            kwargs["cells"] = cell_seq

        edge_seq = None
        for instance in instances.values():
            seq = instance.get_sequence_from_list(self.edges, "edge")
            if cell_seq is None:
                edge_seq = seq
            elif seq is not None:
                edge_seq += seq

        if edge_seq is not None:
            kwargs["edges"] = edge_seq

        assembly.Set(name=self.name, **kwargs)



import abml_helpers
from abml.abml_dataclass import Abml_Registry
from numpy import array


@Abml_Registry.register("sketch")
class Abml_Sketch:
    def __init__(self, model, **kwargs):
        self.model = model
        self.type = kwargs.pop("cmd", "rect")
        self.kwargs = kwargs

        self._sketch_reg = {
            "rect": self._create_sketch_rect,
            "line": self._create_sketch_line,
            "pline": self._create_sketch_pline,
            "circle": self._create_sketch_circle,
            "ellipse": self._create_sketch_ellipse,
            "arc_center_2points": self._create_sketch_arc_c2p,
            "arc_3points": self._create_sketch_arc_3p,
            "autotrim": self._autotrim,
        }

        self.create()

    def create(self):
        self.model.m.ConstrainedSketch(name="__profile__", sheetSize=200.0)
        self._sketch_reg.get(self.type, self._create_sketch_rect)(**self.kwargs)

    def _create_sketch_rect(self, p1, p2):
        s = self.model.m.sketches["__profile__"]
        s.rectangle(point1=p1, point2=p2)

    def _create_sketch_line(self, p1, p2):
        s = self.model.m.sketches["__profile__"]
        s.Line(point1=p1, point2=p2)

    def _create_sketch_pline(self, points):
        s = self.model.m.sketches["__profile__"]
        lines = create_lines_from_points(array(points))
        for line in lines:
            s.Line(point1=line[0], point2=line[1])

    def _create_sketch_circle(self, center, r):
        s = self.model.m.sketches["__profile__"]
        s.CircleByCenterPerimeter(center=center, point1=r)

    def _create_sketch_ellipse(self, center, r1, r2):
        s = self.model.m.sketches["__profile__"]
        s.EllipseByCenterPerimeter(center=center, axisPoint1=r1, axisPoint2=r2)

    def _create_sketch_arc_c2p(self, center, p1, p2, **kwargs):
        s = self.model.m.sketches["__profile__"]

        direction = kwargs.get("direction", "ccw")

        direction_reg = {
            "ccw": "COUNTERCLOCKWISE",
            "counterclockwise": "COUNTERCLOCKWISE",
            "cw": "COUNTERCLOCKWISE",
            "clockwise": "COUNTERCLOCKWISE",
        }

        s.ArcByCenterEnds(center=center, point1=p1, point2=p2, direction=direction_reg[direction.lower()])

    def _create_sketch_arc_3p(self, p1, p2, p3):
        s = self.model.m.sketches["__profile__"]

        s.Arc3Points(point1=p1, point2=p2, point3=p3)

    def _autotrim(self, p1):
        s = self.model.m.sketches["__profile__"]

        geometries = self._point_on_geometry(p1)

        for geometry in geometries:
            s.autoTrimCurve(curve1=geometry, point1=p1, parameter1=None)  # noqq

    def _point_on_geometry(self, p1):
        s = self.model.m.sketches["__profile__"]
        geometries = s.geometry

        geometry_list = []
        for geometry in geometries:  # noqq
            vp1 = geometry.getVertices()[0]
            vp2 = geometry.getVertices()[1]

            if self._point_on_line(p1, vp1, vp2):
                geometry_list.append(geometry)
        return geometry_list

    def _point_on_line(self, p1, vp1, vp2):
        dxc = p1[0] - vp1[0]
        dyc = p1[1] - vp1[1]

        dxl = vp2[0] - vp1[0]
        dyl = vp2[1] - vp1[1]

        cross = dxc * dyl - dyc * dxl

        if cross == 0:
            return True
        return False


def create_lines_from_points(points, lines=[]):
    if len(points) == 0:
        return lines

    if len(lines) == 0:
        if len(points) == 1:
            raise ValueError("Cannot create line from one point")
        return create_lines_from_points(points[2:], [(points[0], points[1])])
    return create_lines_from_points(points[1:], lines + [(lines[-1][1], points[0])])



import abml_helpers
from abaqusConstants import THREE_D, DEFORMABLE_BODY, XZPLANE  # noqa
from abaqusConstants import XYPLANE, YZPLANE, XZPLANE  # noqa
from abaqusConstants import MIDDLE_SURFACE, FROM_SECTION  # noqa
from regionToolset import Region
from abml_dataclass import Abml_Registry
from abml_dataclass import to_object_option_mode, to_object_key_mode, to_object_cmd_mode
from abml_helpers import Abml_Helpers
from abml_sketch import Abml_Sketch
from numpy import array, abs
from abaqus import mdb


@Abml_Registry.register("parts")
class Abml_Part:
    def __init__(self, name, model, **kwargs):  # noqa
        self.model = model
        self.name = name
        self.create()

        geometry = kwargs.get("geometry")
        if geometry is not None:
            params = {
                "data": geometry,
                "type_": "geometry",
                "part": self,
            }
            self.geometry = to_object_option_mode(**params)

        planes = kwargs.get("planes")
        if planes is not None:
            params = {
                "data": planes,
                "type_": "planes",
                "part": self,
            }
            self.planes = to_object_key_mode(**params)

        partitions = kwargs.get("partitions")
        if partitions is not None:
            params = {
                "data": partitions,
                "type_": "partitions",
                "part": self,
            }
            self.partitions = to_object_cmd_mode(**params)

        surfaces = kwargs.get("surfaces")
        if surfaces is not None:
            params = {
                "data": surfaces,
                "type_": "surfaces",
                "part": self,
            }
            self.surfaces = to_object_key_mode(**params)

        sets = kwargs.get("sets")

        if sets is not None:
            params = {
                "data": sets,
                "type_": "sets",
                "part": self,
            }
            self.sets = to_object_key_mode(**params)

        section_assignments = kwargs.get("section_assignments")
        if section_assignments is not None:
            params = {
                "data": section_assignments,
                "type_": "section_assignments",
                "part": self,
            }
            self.section_assignments = to_object_key_mode(**params)

        mesh = kwargs.get("mesh")
        if mesh is not None:
            params = {
                "data": mesh,
                "type_": "mesh",
                "part": self,
            }

            self.mesh = to_object_cmd_mode(**params)
            self.generate_mesh()

    def create(self):
        self.model.m.Part(dimensionality=THREE_D, name=self.name, type=DEFORMABLE_BODY)

    def get_sequence_from_set_list(self, set_list):
        face_seq = None
        cell_seq = None
        edge_seq = None
        for set_ in set_list:
            seq = self.model.p[self.name].sets[set_].faces
            if face_seq is None:
                face_seq = seq
            else:
                face_seq += seq

            seq = self.model.p[self.name].sets[set_].cells
            if cell_seq is None:
                cell_seq = seq
            else:
                cell_seq += seq

            seq = self.model.p[self.name].sets[set_].edges
            if edge_seq is None:
                edge_seq = seq
            else:
                edge_seq += seq

        kwargs = {
            "faces": face_seq,
            "cells": cell_seq,
            "edges": edge_seq,
        }

        return kwargs

    def generate_mesh(self):
        self.model.p[self.name].generateMesh()


@Abml_Registry.register("geometry")
class Abml_Geometry(Abml_Helpers):
    def __init__(self, type, sketch, part, **kwargs):  # noqa
        self.type = type
        self.part = part
        self.model = part.model

        params = {"data": sketch, "type_": "sketch", "model": self.model}

        self.sketch = to_object_cmd_mode(**params)

        self.kwargs = kwargs
        self.map = {
            ("3d", "deformable", "solid_extrusion"): self._create_solid_extrusion,
        }

        self.create()

    def create(self):
        self.map[("3d", "deformable", self.type)]()

    def _create_solid_extrusion(self):
        if not isinstance(self.kwargs["depth"], (int, float)) or self.kwargs["depth"] == 0.0:
            error_str = "Extrusion depth for part: {} is no int, float or is 0".format(self.part.name)
            raise ValueError(error_str)
        part = self.model.p[self.part.name]
        sketch = self.model.m.sketches["__profile__"]
        part.BaseSolidExtrude(depth=self.kwargs["depth"], sketch=sketch)
        del sketch

        self.aliases = {}

    def get_face_sequence(self, element):
        part = self.model.p[self.part.name]
        element = array([element]).flatten()
        seq = None
        if element.shape == (3,) and isinstance(element[0], (float, int)):  # findAt
            face = part.faces.findAt(element.astype(float))  # noqa
            if face is not None:
                seq = Region(side1Faces=part.faces[face.index : face.index + 1]).side1Faces
        elif element.shape == (6,) and isinstance(element[0], (float, int)):  # getByBoundingBox
            faces = part.faces.getByBoundingBox(*element.astype(float))
            if len(faces) != 0:
                seq = faces
        elif isinstance(element.item(0), str):  # getByBoundingBox str
            if any(item == element.item(0) for item in ["all"]):
                seq = part.faces[:]
            elif any(item == element.item(0) for item in part.surfaces.keys()):
                seq = part.surfaces[element.item(0)].faces
            else:
                boundaries = self.get_boundary_rect(self.sketch[0], self.kwargs["depth"], element.item(0))
                seq = part.faces.getByBoundingBox(*boundaries)

        return seq

    def get_cell_sequence(self, element):
        part = self.model.p[self.part.name]
        element = array([element]).flatten()
        seq = None
        if element.shape == (3,) and isinstance(element[0], (float, int)):  # findAt
            cell = part.cells.findAt(element.astype(float))
            if cell is not None:
                seq = Region(cells=part.cells[cell.index : cell.index + 1]).cells
        elif element.shape == (6,) and isinstance(element[0], (float, int)):  # getByBoundingBox
            cells = part.cells.getByBoundingBox(*element.astype(float))
            if len(cells) != 0:
                seq = cells
        elif isinstance(element.item(0), str):  # getByBoundingBox
            if element.item(0) == "all":
                seq = part.cells[:]
            else:
                boundaries = self.get_boundary_rect(self.sketch[0], self.kwargs["depth"], element.item(0))
                seq = part.cells.getByBoundingBox(*boundaries)
        return seq

    def get_edge_sequence(self, element):
        part = self.model.p[self.part.name]
        element = array([element]).flatten()
        seq = None
        if element.shape == (3,) and isinstance(element[0], (float, int)):  # findAt
            edge = part.edges.findAt(element.astype(float))
            if isinstance(edge, type(None)):
                seq = None
            else:
                seq = Region(edges=part.edges[edge.index : edge.index + 1]).edges
        elif element.shape == (6,) and isinstance(element[0], (float, int)):  # getByBoundingBox
            edges = part.edges.getByBoundingBox(*element.astype(float))
            if len(edges) != 0:
                seq = edges
        elif isinstance(element.item(0), str):  # getByBoundingBox
            if any(item == element.item(0) for item in ["all"]):
                seq = part.edges[:]
            else:
                boundaries = self.get_boundary_rect(self.sketch[0], self.kwargs["depth"], element.item(0))
                seq = part.edges.getByBoundingBox(*boundaries)

        return seq

    def get_set_sequence_edges(self, element):
        return Region(edges=self.model.p[self.part.name].sets[element].edges).edges

    def get_set_sequence_cells(self, element):
        return Region(cells=self.model.p[self.part.name].sets[element].cells).cells

    def get_set_sequence_faces(self, element):
        return Region(side1Faces=self.model.p[self.part.name].sets[element].faces).faces

    def get_sequence_from_list(self, elements, type_):
        sequence_map = {
            "cell": self.get_cell_sequence,
            "face": self.get_face_sequence,
            "edge": self.get_edge_sequence,
            "set_edges": self.get_set_sequence_edges,
            "set_faces": self.get_set_sequence_faces,
            "set_cells": self.get_set_sequence_cells,
        }

        type_seq = None
        for element in elements:
            if len(element) != 0:
                seq = sequence_map[type_](element)
                if type_seq is None:
                    type_seq = seq
                else:
                    type_seq += seq
        return type_seq


@Abml_Registry.register("planes")
class Abml_Plane:
    def __init__(self, name, part, **kwargs):  # noqa
        self.name = name
        self.part = part
        self.model = part.model
        self.kwargs = kwargs

        self._principal_Plane = {"x": YZPLANE, "y": XZPLANE, "z": XYPLANE}
        self.create()

    def create(self):
        part = self.model.p[self.part.name]
        for key, value in self.kwargs.items():
            part.DatumPlaneByPrincipalPlane(offset=value, principalPlane=self._principal_Plane[key])
            part.features.changeKey(fromName="Datum plane-1", toName=self.name)  # noqa


@Abml_Registry.register("partitions")
class Abml_Partition(Abml_Helpers):
    def __init__(self, part, faces=None, edges=None, cells=None, **kwargs):
        self.method = kwargs.get("cmd", "plane")
        self.part = part
        self.model = part.model
        self.faces = self.convert_position_list(faces)
        self.edges = self.convert_position_list(edges)
        self.cells = self.convert_position_list(cells)
        self.kwargs = kwargs
        self.partition_map = {"plane": self.create_face_plane}

        self.create()

    def create(self):
        self.partition_map[self.method]()

    def create_face_plane(self):
        part_ = self.model.p[self.part.name]
        planes = array([self.kwargs.get("planes", [])]).flatten()

        for plane in planes:
            id_ = part_.features[plane].id

            if self.faces is not None:
                face_seq = self.part.geometry.get_sequence_from_list(self.faces, "face")
                if face_seq is not None:
                    try:
                        part_.PartitionFaceByDatumPlane(datumPlane=part_.datums[id_], faces=face_seq)
                    except Exception as e:
                        error_msg = """
                        part_name: {part_name}
                        number of face seq found: {num_face_seq}
                        plane: {plane}

                        {e}
                        """.format(
                            part_name=self.part.name,
                            e=e,
                            num_face_seq=len(face_seq),
                            plane=plane,
                        )
                        raise ValueError(error_msg)

            if self.cells is not None:
                cell_seq = self.part.geometry.get_sequence_from_list(self.cells, "cell")
                if cell_seq is not None:
                    part_.PartitionCellByDatumPlane(datumPlane=part_.datums[id_], cells=cell_seq)

            if self.edges is not None:
                edge_seq = self.part.geometry.get_sequence_from_list(self.edges, "edge")
                if edge_seq is not None:
                    part_.PartitionEdgeByDatumPlane(datumPlane=part_.datums[id_], edges=edge_seq)


@Abml_Registry.register("surfaces")
class Abml_Surface(Abml_Helpers):
    def __init__(self, name, faces, part, **kwargs):
        self.name = name
        self.part = part
        self.model = self.part.model
        self.faces = self.convert_position_list(faces)

        self.create()

    def create(self):
        part = self.model.p[self.part.name]

        face_seq = self.part.geometry.get_sequence_from_list(self.faces, "face")

        try:
            part.Surface(side1Faces=face_seq, name=self.name)
        except Exception as e:
            error_msg = """
            surface_name: {name}
            {e}
            """.format(
                name=self.name,
                e=e,
            )
            raise ValueError(error_msg)


@Abml_Registry.register("sets")
class Abml_Sets(Abml_Helpers):
    def __init__(self, name, part, faces=[], edges=[], cells=[]):
        self.name = name
        self.part = part
        self.model = part.model
        self.faces = self.convert_position_list(faces)
        self.edges = self.convert_position_list(edges)
        self.cells = self.convert_position_list(cells)
        self.create()

    def create(self):
        part = self.model.p[self.part.name]

        kwargs = {}
        face_seq = self.part.geometry.get_sequence_from_list(self.faces, "face")
        if face_seq is not None:
            kwargs["faces"] = face_seq

        cell_seq = self.part.geometry.get_sequence_from_list(self.cells, "cell")
        if cell_seq is not None:
            kwargs["cells"] = cell_seq

        edge_seq = self.part.geometry.get_sequence_from_list(self.edges, "edge")
        if edge_seq is not None:
            kwargs["edges"] = edge_seq

        part.Set(name=self.name, **kwargs)


@Abml_Registry.register("section_assignments")
class Abml_Section_Assignments(Abml_Helpers):
    def __init__(self, name, part, cells=None, sets=None):
        self.name = name
        self.part = part
        self.model = part.model

        self.cells = self.convert_position_list(cells)
        self.sets = self.convert_position_list(sets)
        self.create()

    def create(self):
        cell_seq = self.part.geometry.get_sequence_from_list(self.cells, "cell")
        if self.sets is not None:
            set_seq = self.part.get_sequence_from_set_list(self.sets)["cells"]
            if set_seq is not None:
                cell_seq += set_seq

        kwargs = {
            "sectionName": self.name,
            "offset": 0.0,
            "offsetType": MIDDLE_SURFACE,
            "offsetField": "",
            "thicknessAssignment": FROM_SECTION,
            "region": Region(cells=cell_seq),
        }

        self.model.p[self.part.name].SectionAssignment(**kwargs)



from abml_dataclass import Abml_Registry, to_object_key_based_type
from abaqusConstants import UNIFORM, ISOTROPIC


@Abml_Registry.register("materials")
class Abml_Material:
    """
    A class representing a material in an Abaqus/Standard simulation.

    Parameters
    ----------
    model : AbaqusModel
        The Abaqus model to which the material belongs.
    name : str
        The name of the material.
    behaviors : dict
        A dictionary of material behaviors, where the keys are the behavior names
        and the values are the corresponding behavior parameters.
    description : str, optional
        A description of the material.
    **kwargs
        Additional keyword arguments to be passed to the Abaqus API.

    Attributes
    ----------
    model : AbaqusModel
        The Abaqus model to which the material belongs.
    description : str
        A description of the material.
    name : str
        The name of the material.
    behaviors : dict
        A dictionary of material behaviors, where the keys are the behavior names
        and the values are the corresponding behavior parameters.

    Methods
    -------
    create()
        Creates the material in the Abaqus model.

    Notes
    -----
    This class is a wrapper around the Abaqus/Standard API for creating and managing
    materials in a finite element simulation. It provides a more user-friendly
    interface for defining material properties and behaviors.

    Examples
    --------
    # python
    >>> model = AbaqusModel()
    >>> behaviors = {"elastic": {"youngs_modulus": 100e9, "poissons_ratio": 0.3}}
    >>> material = Abml_Material(model, "steel", behaviors, "steel")
    >>> material.create()

    .. code-block:: yaml

        key1: value1
        key2: value2
        key3:
            - item1
            - item2
    """

    def __init__(self, model, name, behaviors, description="", **kwargs):  # noqa
        self.model = model
        self.description = description
        self.name = name
        self.create()
        self.behaviors = to_object_key_based_type(behaviors, material=self)

    def create(self):
        self.model.m.Material(name=self.name, description=self.description)


@Abml_Registry.register("density")
class Abml_Density:
    def __init__(self, density, material, distribution="uniform"):
        self.material = material
        self.distribution = distribution
        self.density = float(density)

        self.create()

    def create(self):
        distribution_type = {"uniform": UNIFORM}
        kwargs = {"table": ((self.density,),), "distributionType": distribution_type[self.distribution]}
        self.material.model.m.materials[self.material.name].Density(**kwargs)


@Abml_Registry.register("elastic")
class Abml_Elastic:
    def __init__(self, E, nu, material, type="isotropic"):  # noqa
        self.material = material
        self.E = float(E)
        self.nu = float(nu)
        self.type = type
        self.types_map = {
            "isotropic": ISOTROPIC,
        }

        self.elastic_map = {"isotropic": self.isotrtopic}

        self.create()

    def create(self):
        self.elastic_map[self.type]()

    def isotrtopic(self):
        kwargs = {"table": ((self.E, self.nu),), "type": self.types_map[self.type]}
        self.material.model.m.materials[self.material.name].Elastic(**kwargs)



from abaqus import mdb
import sys
import os
import re


class IterReg(type):
    # Register of all objects for iteration
    def __iter__(cls):
        return iter(cls._reg.values())


class Abml_Helpers(object):
    @classmethod
    def map_function(cls, key):
        def decorator(func):
            cls.map[key] = func
            return func

        return decorator

    @classmethod
    def get_boundary_rect(cls, sketch, depth, orientation):
        p1 = sketch.kwargs["p1"]
        p2 = sketch.kwargs["p2"]
        x, y, z = abs(p1[0] - p2[0]), abs(p1[1] - p2[1]), depth

        offset = 1e-5
        boundaries = {
            "W": (-offset, 0, 0, 0, y, z),
            "N": (0, -offset, 0, x, 0, z),
            "T": (0, 0, -offset, x, y, 0),
            "top": (0, 0, -offset, x, y, 0),
            "E": (x - offset, 0, 0, x, y, z),
            "S": (0, y - offset, 0, x, y, z),
            "B": (0, 0, z - offset, x, y, z),
            "bot": (0, 0, z - offset, x, y, z),
        }

        return boundaries[orientation]

    @classmethod
    def convert_position_list(cls, position_list):
        if position_list is not None:

            def depth(list_):
                if isinstance(list_, list) and len(list_) != 0:
                    return 1 + max(depth(item) for item in list_)
                return 0

            depth_list = depth(position_list)
            if depth_list == 1 and all(isinstance(x, (int, float)) for x in position_list):
                position_list = [position_list]
            elif depth_list == 0:
                position_list = [position_list]
            return position_list
        return None


def exit_handler():
    for file in os.listdir("."):
        if ".rpy" in file:
            try:
                os.remove(file)
            except WindowsError:
                pass
        if ".rec" in file:
            try:
                os.remove(file)
            except WindowsError:
                pass
        if ".dmp" in file:
            try:
                os.remove(file)
            except WindowsError:
                pass


def cprint(string, prefix="info"):
    print >> sys.__stdout__, prefix + " " + str(string)  # type: ignore # noqa


patter_camel_to_snake = re.compile(r"(?<!^)(?=[A-Z])")
patter_snake_to_camel = re.compile(r"(.*?)_([a-zA-Z])")


def _camel(match):
    return match.group(1) + match.group(2).upper()


def camel_to_snake(name):
    return patter_camel_to_snake.sub("_", name).lower()


def snake_to_camel(name):
    return patter_snake_to_camel.sub(_camel, name, 0)


def get_objects(model_name):
    m = mdb.models[model_name]
    a = m.rootAssembly
    p = m.parts

    return m, a, p



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['abml']

package_data = \
{'': ['*']}

install_requires = \
['pyaml>=21.10.1,<22.0.0']

setup_kwargs = {
    'name': 'abml',
    'version': '0.1.0',
    'description': '',
    'long_description': '',
    'author': 'DavidNaizheZhou',
    'author_email': '70525024+DavidNaizheZhou@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '==2.7.15',
}


setup(**setup_kwargs)



from abml_dataclass import Abml_Registry
from abml_helpers import cprint


@Abml_Registry.register("steps")
class Abml_Step:
    def __init__(self, name, type, pre_step, model):  # noqa
        self.name = name
        self.type = type
        self.model = model
        self.pre_step = pre_step
        self.map = {"static_general": self.static_general}
        self.create()

    def create(self):
        self.map[self.type]()

    def static_general(self):
        self.model.m.StaticStep(name=self.name, previous=self.pre_step)
